from django.apps import apps as django_apps
from django.contrib import admin
from django.utils.translation import gettext_lazy as _
from edc_model_admin.dashboard import ModelAdminSubjectDashboardMixin

from ..admin_site import edc_crf_admin
from ..models import CrfStatus

crf_status_fieldset_tuple = (
    _("CRF status"),
    {"fields": ("crf_status", "crf_status_comments")},
)


@admin.register(CrfStatus, site=edc_crf_admin)
class CrfStatusAdmin(ModelAdminSubjectDashboardMixin, admin.ModelAdmin):
    list_display = (
        "subject_identifier",
        "crf",
        "dashboard",
        "visit",
        "schedule",
        "created",
    )

    list_filter = (
        "created",
        "user_created",
        "user_modified",
        "visit_code",
        "label_lower",
    )

    search_fields = ("subject_identifier",)

    def get_subject_dashboard_url_kwargs(self, obj):
        return dict(
            subject_identifier=obj.subject_identifier,
            visit_schedule_name=obj.visit_schedule_name,
            schedule_name=obj.schedule_name,
            visit_code=obj.visit_code,
        )

    def visit(self, obj):
        return f"{obj.visit_code}.{obj.visit_code_sequence}"

    def crf(self, obj):
        model_cls = django_apps.get_model(obj.label_lower)
        return model_cls._meta.verbose_name

    def schedule(self, obj):
        return f"{obj.visit_schedule_name}.{obj.schedule_name}"



from django import forms
from django.test import TestCase, override_settings
from edc_appointment.models import Appointment
from edc_consent import site_consents
from edc_constants.constants import INCOMPLETE
from edc_facility import import_holidays
from edc_form_validators import FormValidator, FormValidatorMixin
from edc_reference import site_reference_configs
from edc_utils import get_utcnow
from edc_visit_schedule import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED
from edc_visit_tracking.tests.helper import Helper
from visit_schedule_app.consents import v1_consent
from visit_schedule_app.models import SubjectConsent, SubjectScreening, SubjectVisit

from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
from edc_crf.modelform_mixins import CrfModelFormMixin

from ..models import Crf
from ..visit_schedule import visit_schedule


@override_settings(
    SUBJECT_CONSENT_MODEL="visit_schedule_app.subjectconsent",
    SUBJECT_SCREENING_MODEL="visit_schedule_app.subjectscreening",
)
class EdcCrfTestCase(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        SubjectScreening.objects.create(
            screening_identifier="ABCD",
            screening_datetime=get_utcnow(),
            subject_identifier="12345",
        )
        self.subject_identifier = "12345"
        site_consents.registry = {}
        site_consents.register(v1_consent)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            subject_consent_model_cls=SubjectConsent,
            onschedule_model_name="visit_schedule_app.onschedule",
        )
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule)
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_visit_tracking.subjectvisit"}
        )
        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all()[0]
        self.subject_visit = SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )

    def test_form_validator_with_crf(self):
        class MyFormValidator(CrfFormValidatorMixin, FormValidator):
            def clean(self) -> None:
                """test all methods"""
                _ = self.related_visit
                _ = self.subject_consent
                _ = self.subject_identifier
                _ = self.report_datetime
                _ = self.age_in_years
                _ = self.offschedule_datetime

        class MyForm(CrfModelFormMixin, FormValidatorMixin, forms.ModelForm):
            form_validator_cls = MyFormValidator

            def validate_against_consent(self):
                pass

            class Meta:
                model = Crf
                fields = "__all__"

        data = dict(
            report_datetime=self.subject_visit.report_datetime,
            subject_visit=self.subject_visit,
            crf_status=INCOMPLETE,
        )
        form = MyForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)

        # missing report_datetime
        data = dict(
            report_datetime=None,
            subject_visit=self.subject_visit,
            crf_status=INCOMPLETE,
        )
        form = MyForm(data=data)
        form.is_valid()
        self.assertIn("report_datetime", form._errors)

        # gets report_datetime from instance
        data = dict(
            subject_visit=self.subject_visit,
            crf_status=INCOMPLETE,
        )
        form = MyForm(data=data)
        form.is_valid()
        self.assertIn("report_datetime", form._errors)

        # missing related_visit.subject_identifier not possible
        # always returns subject_identifier
        self.subject_visit.subject_identifier = None
        self.subject_visit.save()
        self.subject_visit.refresh_from_db()
        self.assertIsNotNone(self.subject_visit.subject_identifier)
        data = dict(
            report_datetime=self.subject_visit.report_datetime,
            subject_visit=self.subject_visit,
            crf_status=INCOMPLETE,
        )
        form = MyForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)

    def test_form_validator_with_prn(self):
        class MyFormValidator(CrfFormValidatorMixin, FormValidator):
            def clean(self) -> None:
                """test all methods"""
                _ = self.subject_consent
                _ = self.subject_identifier
                _ = self.report_datetime

        class MyForm(CrfModelFormMixin, FormValidatorMixin, forms.ModelForm):
            form_validator_cls = MyFormValidator

            def validate_against_consent(self):
                pass

            class Meta:
                model = Crf
                fields = "__all__"

        data = dict(
            report_datetime=self.subject_visit.report_datetime,
            subject_visit=self.subject_visit,
            crf_status=INCOMPLETE,
        )
        form = MyForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)



from django.db.models.signals import post_save
from django.dispatch import receiver


@receiver(
    post_save,
    weak=False,
    dispatch_uid="update_crf_status_post_save",
)
def update_crf_status_post_save(sender, instance, raw, created, using, **kwargs):
    if not raw and not kwargs.get("update_fields"):
        if ".historical" not in instance._meta.label_lower:
            try:
                instance.update_crf_status_for_instance
            except AttributeError as e:
                if "update_crf_status_for_instance" not in str(e):
                    raise AttributeError(str(e))
            else:
                instance.update_crf_status_for_instance()



from .crf_model_form_mixin import CrfModelFormMixin
from .crf_singleton_model_form_mixin import CrfSingletonModelFormMixin
from .inline_model_form_mixin import InlineCrfModelFormMixin
from .requisition_model_form_mixin import RequisitionModelFormMixin



from typing import Iterator, List, Type, TypeVar, Union

from django.db import models
from edc_protocol import Protocol
from edc_visit_tracking.stubs import SubjectVisitModelStub


class MetaModelStub(Protocol):
    verbose_name: str
    fields: Iterator
    ...


class CrfModelStub(Protocol):
    subject_visit: SubjectVisitModelStub

    @classmethod
    def related_visit_model_attr(cls) -> str:
        ...

    objects: models.Manager
    _meta: MetaModelStub
    ...


TCrfModelStub = TypeVar("TCrfModelStub", bound="CrfModelStub")


class MetaModelFormStub(Protocol):
    model: Type[CrfModelStub]
    fields: Union[str, List[str]]
    ...


class CrfModelFormStub(Protocol):
    cleaned_data: dict
    instance: CrfModelStub
    _meta: MetaModelFormStub
    ...



from __future__ import annotations

from typing import TYPE_CHECKING

from edc_visit_tracking.modelform_mixins import get_related_visit

if TYPE_CHECKING:
    from edc_visit_tracking.model_mixins import VisitModelMixin


class InlineCrfModelFormMixin:
    @property
    def related_visit(self) -> VisitModelMixin:
        """Return the instance of the inline parent model's visit
        model.
        """
        return get_related_visit(
            self,
            related_visit_model_attr=self.instance.parent_model.related_visit_model_attr(),
        )

    @property
    def related_visit_model_attr(self) -> str:
        return self.instance.parent_model.related_visit_model_attr()



from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase, override_settings
from edc_appointment.models import Appointment
from edc_consent import site_consents
from edc_constants.constants import COMPLETE, INCOMPLETE
from edc_facility import import_holidays
from edc_reference import site_reference_configs
from edc_utils import get_utcnow
from edc_visit_schedule import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED
from edc_visit_tracking.tests.helper import Helper
from visit_schedule_app.consents import v1_consent
from visit_schedule_app.models import SubjectConsent, SubjectVisit

from edc_crf.models import CrfStatus

from ..models import Crf
from ..visit_schedule import visit_schedule


@override_settings(
    SUBJECT_CONSENT_MODEL="visit_schedule_app.subjectconsent",
    SUBJECT_SCREENING_MODEL="visit_schedule_app.subjectscreening",
)
class EdcCrfTestCase(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        self.subject_identifier = "12345"
        site_consents.registry = {}
        site_consents.register(v1_consent)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            subject_consent_model_cls=SubjectConsent,
            onschedule_model_name="visit_schedule_app.onschedule",
        )
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule)
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_visit_tracking.subjectvisit"}
        )
        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all()[0]
        self.subject_visit = SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )

    def test_default_incomplete(self):
        crf_obj = Crf.objects.create(subject_visit=self.subject_visit)
        self.assertEqual(crf_obj.crf_status, INCOMPLETE)

    @override_settings(CRF_STATUS_DEFAULT=COMPLETE)
    def test_default_complete(self):
        crf_obj = Crf.objects.create(subject_visit=self.subject_visit)
        self.assertEqual(crf_obj.crf_status, COMPLETE)

    def test_creates_crf_status(self):
        crf_obj = Crf.objects.create(subject_visit=self.subject_visit)
        try:
            CrfStatus.objects.get(
                subject_identifier=crf_obj.subject_visit.subject_identifier,
                label_lower="edc_crf.crf",
            )
        except ObjectDoesNotExist:
            self.fail("crf status unexpectedly does not exist")

    def test_deletes_crf_status(self):
        crf_obj = Crf.objects.create(subject_visit=self.subject_visit)
        self.assertEqual(crf_obj.crf_status, INCOMPLETE)
        crf_obj.crf_status = COMPLETE
        crf_obj.save()
        self.assertRaises(
            ObjectDoesNotExist,
            CrfStatus.objects.get,
            subject_identifier=crf_obj.subject_visit.subject_identifier,
            label_lower="edc_crf.crf",
        )



from edc_action_item.models import ActionNoManagersModelMixin
from edc_model.models import HistoricalRecords
from edc_sites.models import SiteModelMixin
from edc_visit_tracking.managers import CrfCurrentSiteManager, CrfModelManager

from .crf_model_mixin import CrfNoManagerModelMixin


class CrfWithActionModelMixin(
    SiteModelMixin,
    CrfNoManagerModelMixin,
    ActionNoManagersModelMixin,
):
    action_name = None

    objects = CrfModelManager()
    on_site = CrfCurrentSiteManager()
    history = HistoricalRecords(inherit=True)

    class Meta(CrfNoManagerModelMixin.Meta):
        abstract = True
